"""
***************
loads
***************

.. currentmodule:: compas_fd.loads

Classes
=======

.. autosummary::
    :toctree: generated/
    :nosignatures:

    SelfweightCalculator

"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from .selfweight import SelfweightCalculator

__all__ = ["SelfweightCalculator"]



from numpy import array
from numpy import asarray
from numpy import float64

from compas_fd.loads import SelfweightCalculator
from compas_fd.fd.fd_constrained_numpy import fd_constrained_numpy
from compas_fd.datastructures import CableMesh


def mesh_fd_constrained_numpy(
    mesh: CableMesh,
    kmax: int = 100,
    damping: float = 0.1,
    tol_res: float = 1e-3,
    tol_disp: float = 1e-3,
) -> CableMesh:
    """
    Iteratively find the equilibrium shape of a mesh for the given force densities.

    Parameters
    ----------
    mesh : :class:`compas_fd.datastructures.CableMesh`
        The mesh to equilibriate.

    Returns
    -------
    :class:`compas_fd.datastructures.CableMesh`
        The function updates the mesh in place,
        but returns a reference to the updated mesh as well
        for compatibility with RPCs.

    """
    v_i = mesh.vertex_index()
    vertices = array(mesh.vertices_attributes("xyz"), dtype=float64)
    fixed = [v_i[v] for v in mesh.vertices_where({"is_anchor": True})]
    edges = list(mesh.edges_where(_is_edge=True))
    qs = asarray(mesh.edges_attribute("q", keys=edges), dtype=float64).reshape((-1, 1))
    loads = array(mesh.vertices_attributes(("px", "py", "pz")), dtype=float64)
    constraints = list(mesh.vertices_attribute("constraint"))
    selfweight = SelfweightCalculator(
        mesh,
        density=mesh.attributes["density"],
        thickness_attr_name="t",
    )

    ij = [(v_i[u], v_i[v]) for u, v in edges]

    result = fd_constrained_numpy(
        vertices=vertices,
        fixed=fixed,
        edges=ij,
        forcedensities=qs,
        loads=loads,
        constraints=constraints,
        kmax=kmax,
        tol_res=tol_res,
        tol_disp=tol_disp,
        damping=damping,
        selfweight=selfweight,
    )

    _update_mesh(mesh, result)

    return mesh


def _update_mesh(mesh, result):
    vertex_index = mesh.vertex_index()
    for vertex, attr in mesh.vertices(True):
        index = vertex_index[vertex]
        attr["x"] = result.vertices[index, 0]
        attr["y"] = result.vertices[index, 1]
        attr["z"] = result.vertices[index, 2]
        attr["_rx"] = result.residuals[index, 0]
        attr["_ry"] = result.residuals[index, 1]
        attr["_rz"] = result.residuals[index, 2]

    for index, (vertex, attr) in enumerate(mesh.edges_where({"_is_edge": True}, True)):
        attr["_f"] = result.forces[index, 0]
        attr["_l"] = result.lenghts[index, 0]



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from compas.geometry import vector_component
from compas.geometry import Vector
from compas.geometry import Frame
from .constraint import Constraint


class FrameConstraint(Constraint):
    def __init__(self, frame, **kwargs):
        super(FrameConstraint, self).__init__(geometry=frame, **kwargs)

    @property
    def data(self):
        return {
            "geometry": self.geometry.data,
            "rhino_guid": str(self._rhino_guid),
        }

    @data.setter
    def data(self, data):
        self.geometry = Frame.from_data(data["geometry"])
        self._rhino_guid = str(data["rhino_guid"])

    @classmethod
    def from_data(cls, data):
        frame = Frame.from_data(data["geometry"])
        constraint = cls(frame)
        constraint._rhino_guid = str(data["rhino_guid"])
        return constraint

    def compute_tangent(self):
        self._tangent = self.residual - self.normal

    def compute_normal(self):
        normal = self.geometry.zaxis
        self._normal = Vector(*vector_component(self.residual, normal))

    def update(self, damping=0.1):
        self._location = self.location + self.tangent * damping

    def project(self):
        pass



"""
***************************
datastructures
***************************

.. currentmodule:: compas_fd.datastructures

Classes
=======

.. autosummary::
    :toctree: generated/
    :nosignatures:

    CableMesh

"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from .cablemesh import CableMesh

__all__ = ["CableMesh"]



from typing import Tuple
from typing import List
from typing import Union
from typing import Sequence
from typing import Optional
from typing_extensions import Literal
from typing_extensions import Annotated
from nptyping import NDArray
from nptyping import Float64

from numpy import asarray
from numpy import zeros_like
from numpy import float64
from scipy.sparse import diags
from scipy.sparse.linalg import spsolve

from compas.numerical import connectivity_matrix
from compas.numerical import normrow

from compas_fd.fd.result import Result


FloatNx3 = Union[
    Sequence[Annotated[List[float], 3]],
    NDArray[Literal["*, 3"], Float64],
]


def fd_numpy(
    *,
    vertices: FloatNx3,
    fixed: List[int],
    edges: List[Tuple[int, int]],
    forcedensities: List[float],
    loads: Optional[FloatNx3] = None,
) -> Result:
    """
    Compute the equilibrium coordinates of a system of vertices connected by edges.
    """
    v = len(vertices)
    free = list(set(range(v)) - set(fixed))
    xyz = asarray(vertices, dtype=float64).reshape((-1, 3))
    q = asarray(forcedensities, dtype=float64).reshape((-1, 1))
    if loads is None:
        p = zeros_like(xyz)
    else:
        p = asarray(loads, dtype=float64).reshape((-1, 3))
    C = connectivity_matrix(edges, "csr")
    Ci = C[:, free]
    Cf = C[:, fixed]
    Ct = C.transpose()
    Cit = Ci.transpose()
    Q = diags([q.flatten()], [0])
    A = Cit.dot(Q).dot(Ci)
    b = p[free] - Cit.dot(Q).dot(Cf).dot(xyz[fixed])
    xyz[free] = spsolve(A, b)
    lengths = normrow(C.dot(xyz))
    forces = q * lengths
    residuals = p - Ct.dot(Q).dot(C).dot(xyz)
    return Result(xyz, residuals, forces, lengths)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import inspect

from compas.data import Data
from compas.geometry import Vector
from compas.geometry import Point

from .exceptions import GeometryNotRegisteredAsConstraint


class Constraint(Data):
    """
    Base class for all constraints.
    """

    GEOMETRY_CONSTRAINT = {}

    @staticmethod
    def get_constraint_cls(geometry, **kwargs):
        gtype = type(geometry)
        cls = None
        for type_ in inspect.getmro(gtype):
            cls = Constraint.GEOMETRY_CONSTRAINT.get(type_)
            if cls is not None:
                break
        if cls is None:
            raise GeometryNotRegisteredAsConstraint(
                "No constraint is registered for this geometry type: {}".format(gtype)
            )
        return cls

    def __new__(cls, *args, **kwargs):
        geometry = args[0]
        cls = Constraint.get_constraint_cls(geometry)
        return super(Constraint, cls).__new__(cls)

    def __init__(self, geometry, **kwargs):
        super(Constraint, self).__init__(**kwargs)
        self._geometry = None
        self._location = None
        self._residual = None
        self._tangent = None
        self._normal = None
        self._rhino_guid = None
        self._param = None
        self.geometry = geometry

    @staticmethod
    def register(gtype, ctype):
        Constraint.GEOMETRY_CONSTRAINT[gtype] = ctype

    @property
    def geometry(self):
        return self._geometry

    @geometry.setter
    def geometry(self, geometry):
        self._residual = None
        self._tangent = None
        self._normal = None
        self._geometry = geometry

    @property
    def location(self):
        return self._location

    @location.setter
    def location(self, point):
        self._tangent = None
        self._normal = None
        self._location = Point(*point)
        if not getattr(self, "projected", None):
            self.project()
            self.projected = True

    @property
    def residual(self):
        return self._residual

    @residual.setter
    def residual(self, residual):
        self._tangent = None
        self._normal = None
        self._residual = Vector(*residual)

    def compute_tangent(self):
        raise NotImplementedError

    def compute_normal(self):
        raise NotImplementedError

    def update(self):
        raise NotImplementedError

    def compute_param(self):
        raise NotImplementedError

    def project(self):
        raise NotImplementedError

    def update_location_at_param(self):
        raise NotImplementedError

    @property
    def tangent(self):
        if self._tangent is None:
            self.compute_tangent()
        return self._tangent

    @property
    def normal(self):
        if self._normal is None:
            self.compute_normal()
        return self._normal

    @property
    def param(self):
        if self._param is None:
            self.compute_param()
        return self._param



"""
******************
fd
******************

.. currentmodule:: compas_fd.fd

Functions
=========

.. autosummary::
    :toctree: generated/
    :nosignatures:

    fd_numpy
    fd_constrained_numpy
    mesh_fd_numpy
    mesh_fd_constrained_numpy

"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from .fd_numpy import fd_numpy
from .fd_constrained_numpy import fd_constrained_numpy
from .mesh_fd_numpy import mesh_fd_numpy
from .mesh_fd_constrained_numpy import mesh_fd_constrained_numpy

__all__ = [
    "fd_numpy",
    "fd_constrained_numpy",
    "mesh_fd_numpy",
    "mesh_fd_constrained_numpy",
]



from numpy import array
from numpy import zeros

from compas.geometry import cross_vectors
from compas.geometry import length_vector
from compas.numerical import face_matrix


class SelfweightCalculator:
    def __init__(self, mesh, density=1.0, thickness_attr_name="t"):
        self.mesh = mesh
        self.rho = array(
            [t * density for t in mesh.vertices_attribute(thickness_attr_name)]
        ).reshape((-1, 1))
        self.key_index = mesh.key_index()
        self.fkey_index = {fkey: index for index, fkey in enumerate(mesh.faces())}
        self.is_loaded = {fkey: True for fkey in mesh.faces()}
        self.F = self.face_matrix()

    def __call__(self, xyz):
        ta = self._tributary_areas(xyz)
        return ta * self.rho

    def face_matrix(self):
        face_vertices = [None] * self.mesh.number_of_faces()
        for fkey in self.mesh.faces():
            face_vertices[self.fkey_index[fkey]] = [
                self.key_index[key] for key in self.mesh.face_vertices(fkey)
            ]
        return face_matrix(face_vertices, rtype="csr", normalize=True)

    def _tributary_areas(self, xyz):
        mesh = self.mesh
        key_index = self.key_index
        fkey_index = self.fkey_index
        is_loaded = self.is_loaded

        C = self.F.dot(xyz)

        areas = zeros((xyz.shape[0], 1))
        for u in mesh.vertices():
            p0 = xyz[key_index[u]]

            a = 0
            for v in mesh.halfedge[u]:
                p1 = xyz[key_index[v]]
                p01 = p1 - p0

                fkey = mesh.halfedge[u][v]
                if fkey is not None and is_loaded[fkey]:
                    p2 = C[fkey_index[fkey]]
                    a += 0.25 * length_vector(cross_vectors(p01, p2 - p0))

                fkey = mesh.halfedge[v][u]
                if fkey is not None and is_loaded[fkey]:
                    p3 = C[fkey_index[fkey]]
                    a += 0.25 * length_vector(cross_vectors(p01, p3 - p0))

            areas[key_index[u]] = a

        return areas



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from compas.geometry import vector_component
from compas.geometry import Vector
from compas.geometry import Point
from compas.geometry import Circle

from .constraint import Constraint


class CircleConstraint(Constraint):
    def __init__(self, curve, **kwargs):
        super(CircleConstraint, self).__init__(geometry=curve, **kwargs)

    @property
    def data(self):
        return {
            "geometry": self.geometry.data,
            "rhino_guid": str(self._rhino_guid),
        }

    @data.setter
    def data(self, data):
        self.geometry = Circle.from_data(data["geometry"])
        self._rhino_guid = str(data["rhino_guid"])

    @classmethod
    def from_data(cls, data):
        curve = Circle.from_data(data["geometry"])
        constraint = cls(curve)
        constraint._rhino_guid = str(data["rhino_guid"])
        return constraint

    @property
    def location(self):
        return self._location

    @location.setter
    def location(self, point):
        self._tangent = None
        self._normal = None
        self._location = Point(*point)
        self.project()

    def compute_tangent(self):
        direction = self.geometry.tangent_at(self.location)
        self._tangent = Vector(*vector_component(self.residual, direction))

    def compute_normal(self):
        self._normal = self.residual - self.tangent

    def update(self, damping=0.1):
        self._location = self.location + self.tangent * damping
        point = self.geometry.closest_point(self.location)
        if self._location.distance_to_point(point) > 0.001:
            self.project()

    def project(self):
        point = self.geometry.closest_point(self.location)
        self._location = point



from typing import Tuple
from typing import List
from typing import Union
from typing import Sequence
from typing import Optional
from typing_extensions import Literal
from typing_extensions import Annotated
from nptyping import NDArray
from nptyping import Float64
from nptyping import Int32

from dataclasses import dataclass
from dataclasses import astuple

from numpy import asarray
from numpy import float64
from numpy import zeros_like
from scipy.sparse import diags

from compas.numerical import connectivity_matrix

from compas_fd.fd.result import Result


@dataclass
class FDNumericalData:
    """
    Stores numerical data used by the force density algorithms.
    """

    free: int
    fixed: int
    xyz: NDArray[Literal["*, 3"], Float64]
    C: NDArray[Literal["*, *"], Int32]
    q: NDArray[Literal["*, 1"], Float64]
    Q: NDArray[Literal["*, *"], Float64]
    p: NDArray[Literal["*, 1"], Float64]
    A: NDArray[Literal["*, *"], Float64]
    Ai: NDArray[Literal["*, *"], Float64]
    Af: NDArray[Literal["*, *"], Float64]
    forces: NDArray[Literal["*, 1"], Float64] = None
    lengths: NDArray[Literal["*, 1"], Float64] = None
    residuals: NDArray[Literal["*, 3"], Float64] = None
    tangent_residuals: NDArray[Literal["*, 3"], Float64] = None
    normal_residuals: NDArray[Literal["*, 1"], Float64] = None

    def __iter__(self):
        return iter(astuple(self))

    @classmethod
    def from_params(
        cls,
        vertices: Union[
            Sequence[Annotated[List[float], 3]],
            NDArray[Literal["*, 3"], Float64],
        ],
        fixed: List[int],
        edges: List[Tuple[int, int]],
        forcedensities: List[float],
        loads: Optional[
            Union[
                Sequence[Annotated[List[float], 3]],
                NDArray[Literal["*, 3"], Float64],
            ]
        ] = None,
    ):
        """
        Construct numerical arrays from force density solver input parameters.
        """
        free = list(set(range(len(vertices))) - set(fixed))
        xyz = asarray(vertices, dtype=float64).reshape((-1, 3))
        C = connectivity_matrix(edges, "csr")
        Ci = C[:, free]
        Cf = C[:, fixed]
        q = asarray(forcedensities, dtype=float64).reshape((-1, 1))
        Q = diags([q.flatten()], [0])
        p = (
            zeros_like(xyz)
            if loads is None
            else asarray(loads, dtype=float64).reshape((-1, 3))
        )
        A = C.T.dot(Q).dot(C)
        Ai = Ci.T.dot(Q).dot(Ci)
        Af = Ci.T.dot(Q).dot(Cf)
        return cls(free, fixed, xyz, C, q, Q, p, A, Ai, Af)

    @classmethod
    def from_mesh(cls, mesh):
        """
        Construct numerical arrays from input mesh.
        """
        raise NotImplementedError

    def to_result(self) -> Result:
        """
        Parse relevant numerical data into a Result object.
        """
        return Result(self.xyz, self.residuals, self.forces, self.lengths)



"""
***************************
constraints
***************************

.. currentmodule:: compas_fd.constraints

Base Class
==========

.. autosummary::
    :toctree: generated/
    :nosignatures:

    Constraint

Classes
=======

.. autosummary::
    :toctree: generated/
    :nosignatures:

    CircleConstraint
    CurveConstraint
    FrameConstraint
    LineConstraint
    PlaneConstraint
    SurfaceConstraint
    VectorConstraint

"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from compas.geometry import Vector
from compas.geometry import Frame
from compas.geometry import Line
from compas.geometry import Plane
from compas.geometry import Circle
from compas.geometry import NurbsCurve
from compas.geometry import NurbsSurface

from .constraint import Constraint

from .vectorconstraint import VectorConstraint
from .frameconstraint import FrameConstraint
from .lineconstraint import LineConstraint
from .planeconstraint import PlaneConstraint
from .circleconstraint import CircleConstraint

from .curveconstraint import CurveConstraint
from .surfaceconstraint import SurfaceConstraint

Constraint.register(Vector, VectorConstraint)
Constraint.register(Frame, FrameConstraint)
Constraint.register(Line, LineConstraint)
Constraint.register(Plane, PlaneConstraint)
Constraint.register(Circle, CircleConstraint)

Constraint.register(NurbsCurve, CurveConstraint)
Constraint.register(NurbsSurface, SurfaceConstraint)



"""
********************************************************************************
compas_fd
********************************************************************************

.. currentmodule:: compas_fd


.. toctree::
    :maxdepth: 1

    compas_fd.constraints
    compas_fd.datastructures
    compas_fd.fd
    compas_fd.loads

"""

from __future__ import print_function

import os
import compas


__author__ = ['tom van mele <van.mele@arch.ethz.ch>']
__copyright__ = 'Block Research Group - ETH Zurich'
__license__ = 'MIT License'
__email__ = 'van.mele@arch.ethz.ch'
__version__ = '0.4.0'


get = compas.get


HERE = os.path.dirname(__file__)

HOME = os.path.abspath(os.path.join(HERE, '../../'))
DATA = os.path.abspath(os.path.join(HOME, 'data'))
DOCS = os.path.abspath(os.path.join(HOME, 'docs'))
TEMP = os.path.abspath(os.path.join(HOME, 'temp'))


__all__ = ['HOME', 'DATA', 'DOCS', 'TEMP', 'get']

__all_plugins__ = [
    'compas_fd.install'
]



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from compas.geometry import vector_component
from compas.geometry import is_point_on_segment
from compas.geometry import closest_point_on_segment
from compas.geometry import Vector
from compas.geometry import Point
from compas.geometry import Line
from .constraint import Constraint


class LineConstraint(Constraint):
    def __init__(self, line, **kwargs):
        super(LineConstraint, self).__init__(geometry=line, **kwargs)

    @property
    def data(self):
        return {
            "geometry": self.geometry.data,
            "rhino_guid": str(self._rhino_guid),
        }

    @data.setter
    def data(self, data):
        self.geometry = Line.from_data(data["geometry"])
        self._rhino_guid = str(data["rhino_guid"])

    @classmethod
    def from_data(cls, data):
        line = Line.from_data(data["geometry"])
        constraint = cls(line)
        constraint._rhino_guid = str(data["rhino_guid"])
        return constraint

    def compute_tangent(self):
        direction = self.geometry.direction
        self._tangent = Vector(*vector_component(self.residual, direction))

    def compute_normal(self):
        self._normal = self.residual - self.tangent

    def update(self, damping=0.1):
        self._location = self.location + self.tangent * damping
        if not is_point_on_segment(point=self._location, segment=self._geometry):
            self.project()

    def project(self):
        self._location = Point(*closest_point_on_segment(self._location, self.geometry))

    def compute_param(self):
        v = Vector.from_start_end(self._geometry.start, self._location)
        u = self._geometry.vector
        d = v.dot(u)
        if not d == 0:
            d = d / abs(d)
        self._param = v.length * d

    def update_location_at_param(self):
        d = self._geometry.direction
        self._location = self._geometry.start + d * self._param



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from compas.datastructures import Mesh


class CableMesh(Mesh):
    """
    Extension of the mesh datastructure with attributes and methods related to form finding of tensile surface structures.
    """

    def __init__(self):
        super(CableMesh, self).__init__()
        self.attributes.update(
            {
                "density": 22.0,
            }
        )
        self.default_vertex_attributes.update(
            {
                "x": 0.0,
                "y": 0.0,
                "z": 0.0,
                "px": 0.0,
                "py": 0.0,
                "pz": 0.0,
                "t": 0.0,
                "is_anchor": False,
                "is_fixed": False,
                "constraint": None,
                "param": None,
                "_rx": 0.0,
                "_ry": 0.0,
                "_rz": 0.0,
            }
        )
        self.default_edge_attributes.update(
            {
                "q": 1.0,
                "_is_edge": True,
                "_l": 0.0,
                "_f": 0.0,
                "_r": 0.0,
            }
        )
        self.default_face_attributes.update(
            {
                "is_loaded": True,
            }
        )



from numpy import array
from numpy import float64

from compas_fd.loads import SelfweightCalculator
from compas_fd.fd.fd_numpy import fd_numpy
from compas_fd.datastructures import CableMesh


def mesh_fd_numpy(mesh: CableMesh) -> CableMesh:
    """
    Find the equilibrium shape of a mesh for the given force densities.

    Parameters
    ----------
    mesh : :class:`compas_fd.datastructures.CableMesh`
        The mesh to equilibriate.

    Returns
    -------
    :class:`compas_fd.datastructures.CableMesh`
        The function updates the mesh in place,
        but returns a reference to the updated mesh as well
        for compatibility with RPCs.

    """
    k_i = mesh.key_index()
    fixed = mesh.vertices_where({"is_anchor": True})
    fixed = [k_i[key] for key in fixed]
    xyz = array(mesh.vertices_attributes("xyz"), dtype=float64)
    p = array(mesh.vertices_attributes(("px", "py", "pz")), dtype=float64)
    edges = [(k_i[u], k_i[v]) for u, v in mesh.edges_where({"_is_edge": True})]
    q = array(
        [attr["q"] for key, attr in mesh.edges_where({"_is_edge": True}, True)],
        dtype=float64,
    ).reshape((-1, 1))
    density = mesh.attributes["density"]
    calculate_sw = SelfweightCalculator(mesh, density=density)
    p[:, 2] -= calculate_sw(xyz)[:, 0]

    result = fd_numpy(vertices=xyz, fixed=fixed, edges=edges, forcedensities=q, loads=p)

    for key, attr in mesh.vertices(True):
        index = k_i[key]
        attr["x"] = result.vertices[index, 0]
        attr["y"] = result.vertices[index, 1]
        attr["z"] = result.vertices[index, 2]
        attr["_rx"] = result.residuals[index, 0]
        attr["_ry"] = result.residuals[index, 1]
        attr["_rz"] = result.residuals[index, 2]

    for index, (key, attr) in enumerate(mesh.edges_where({"_is_edge": True}, True)):
        attr["_f"] = result.forces[index, 0]
        attr["_l"] = result.lenghts[index, 0]

    return mesh



from typing import List
from typing import NamedTuple
from typing_extensions import Literal
from nptyping import NDArray

import numpy as np


class Result(NamedTuple):
    vertices: NDArray[Literal["N", 3], np.float64]
    residuals: NDArray[Literal["N", 3], np.float64]
    forces: List[float]
    lenghts: List[float]



from typing import Callable
from typing import Tuple
from typing import List
from typing import Union
from typing import Sequence
from typing import Optional
from typing_extensions import Literal
from typing_extensions import Annotated
from nptyping import NDArray
from nptyping import Float64

from numpy import asarray
from scipy.linalg import norm
from scipy.sparse.linalg import spsolve

from compas.numerical import normrow

from compas_fd.constraints import Constraint
from compas_fd.fd.fd_numerical_data import FDNumericalData
from compas_fd.fd.result import Result


FloatNx3 = Union[
    Sequence[Annotated[List[float], 3]],
    NDArray[Literal["*, 3"], Float64],
]


def fd_constrained_numpy(
    *,
    vertices: FloatNx3,
    fixed: List[int],
    edges: List[Tuple[int, int]],
    forcedensities: List[float],
    loads: Optional[FloatNx3] = None,
    constraints: Sequence[Constraint],
    kmax: int = 100,
    tol_res: float = 1e-3,
    tol_disp: float = 1e-3,
    damping: float = 0.1,
    selfweight=None,
) -> Result:
    """
    Iteratively compute the equilibrium coordinates of a system of vertices connected by edges.
    Vertex constraints are recomputed at each iteration.
    """
    nd = FDNumericalData.from_params(vertices, fixed, edges, forcedensities, loads)

    for k in range(kmax):
        xyz_prev = nd.xyz
        _solve_fd(nd, selfweight)
        _update_constraints(nd, constraints, damping)
        if _is_converged_residuals(
            nd.tangent_residuals, tol_res
        ) and _is_converged_disp(xyz_prev, nd.xyz, tol_disp):
            break

    _post_process_fd(nd)
    return nd.to_result()


def _solve_fd(numdata: FDNumericalData, selfweight: Callable = None) -> None:
    """
    Solve a single iteration for the equilibrium coordinates of a system.
    All updated numerical arrays are stored in the numdata parameter.
    """
    nd = numdata

    p = nd.p.copy()
    if selfweight:
        p[:, 2] -= selfweight(nd.xyz)[:, 0]

    p = p[nd.free]
    b = p - nd.Af.dot(nd.xyz[nd.fixed])
    nd.xyz[nd.free] = spsolve(nd.Ai, b)
    numdata.residuals = nd.p - nd.A.dot(nd.xyz)


def _post_process_fd(numdata: FDNumericalData) -> None:
    """
    Compute dependent numerical arrays from the numerical data after running solver.
    All updated numerical arrays are stored in the numdata parameter.
    """
    nd = numdata
    nd.lengths = normrow(nd.C.dot(nd.xyz))
    numdata.forces = nd.q * nd.lengths


def _update_constraints(
    numdata: FDNumericalData, constraints: Sequence[Constraint], damping: float
) -> None:
    """
    Update all vertex constraints by the residuals of the current iteration,
    and store their updated vertex coordinates in the numdata parameter.
    """
    nd = numdata
    for vertex, constraint in enumerate(constraints):
        if not constraint:
            continue
        constraint.location = nd.xyz[vertex]
        constraint.residual = nd.residuals[vertex]
        constraint.update(damping=damping)
        nd.xyz[vertex] = constraint.location
    nd.tangent_residuals = asarray([c.tangent for c in constraints if c])


def _is_converged_residuals(
    residuals: NDArray[Literal["*, 3"], Float64], tol_res: float
) -> bool:
    """
    Verify whether the maximum constraint residual is within tolerance.
    """
    if residuals is None or not residuals.any():
        return True
    max_res = max(norm(residuals, axis=1))
    return max_res < tol_res


def _is_converged_disp(
    old_xyz: NDArray[Literal["*, 3"], Float64],
    new_xyz: NDArray[Literal["*, 3"], Float64],
    tol_disp: float,
) -> bool:
    """
    Verify whether the maximum coordinate displacement
    between consecutive iterations is within tolerance.
    """
    max_dxyz = max(norm(new_xyz - old_xyz, axis=1))
    return max_dxyz < tol_disp
